"""
Sphinx role
"""
from pathlib import Path

from docutils import nodes, utils
from sphinx.util.nodes import split_explicit_title

from .exceptions import ParseError
from .parser import python_to_lineno


def gitref(name, rawtext, text, lineno, inliner, options={}, content=[]):
    """
    Reference a file in git

    Uses the ``gitref_repo`` config value

    Usage:
        :gitref:`target`
        :gitref:`label <target>`

    where ``target`` is a filename with an optional reference to a definition in the
    code:

    * ``path/to/filename.py``
    * ``path/to/filename.py#L6``
    * ``path/to/filename.py::coderef``
    * ``path/to/filename.py::code.ref``

    Renders HTML:
        <a href="url">label</a>
    """
    # Collect config vars
    app = inliner.document.settings.env.app
    try:
        remote = app.config.gitref_remote
        if not remote:
            raise AttributeError
    except AttributeError:
        raise ValueError("Config does not specify gitref_remote")

    # Assume the project root is one dir up from the docs dir
    doc_root = inliner.document.settings.env.srcdir
    project_root = Path(doc_root).parent

    # Process text value
    has_t, title, target = split_explicit_title(text)
    title = utils.unescape(title)
    target = utils.unescape(target)

    # Break target
    if "::" in target:
        filename, coderef = target.split("::", 1)
    else:
        filename = target
        coderef = None

    # Set title if not set
    if title == target:
        if coderef is not None:
            title = app.config.gitref_label_format.format(
                filename=filename,
                coderef=coderef,
            )

    # Ensure the file exists - can be a file or a dir
    filepath = project_root / filename
    ref_start = None
    if not filepath.exists():
        inliner.reporter.error(f"Referenced file does not exist: {filename}")

    # Convert a code ref into a line number
    elif coderef is not None:
        try:
            ref_start = python_to_lineno(filepath, coderef)
        except ParseError as error:
            inliner.reporter.error(
                f'Error resolving code reference "{target}": {error}',
                line=lineno,
            )
            target = filename

    ref = remote.get_url(filename=filename, line=ref_start)

    node = nodes.reference(rawtext, title, refuri=ref, **options)
    return [node], []



"""
Constant values
"""

#: string: Name of remote git source to look for in current repository
DEFAULT_REMOTE = "origin"

#: string: Format for label if not set, and if a coderef is provided. Available vars:
#   filename    Filename component
#   coderef     Code reference
DEFAULT_LABEL_FORMAT = "{coderef}"



"""
Parse python source
"""
import ast

from .exceptions import NodeNotFound, ParseError


# Prep assignment node types:
#   ast.Assign: assignment, eg ``x = 1``
#   ast.AugAssign: augmented assignment, eg ``x += 1``
#   ast.AnnAssign: annotated assignment, eg ``x: int = 1``
AssignTypes = (ast.Assign, ast.AugAssign, ast.AnnAssign)


def find_name_in_nodes(name, nodes, path=""):
    """
    Given a list of AST nodes, find the named object

    Arguments:
        name: object name, or Python path to object name

    Returns:
        node: AST node that defines the name

    Raises:
        NodeNotFound: no AST node was found matching that name
    """
    # Look for a path
    if "." in name:
        name, rest = name.split(".", 1)
    else:
        rest = ""

    # Track where we came from for reporting purposes
    if not path:
        path = name
    else:
        path = f"{path}.{name}"

    for node in nodes:
        # Check for assignments
        if type(node) in AssignTypes:
            for target in node.targets:
                if not hasattr(target, "id") or target.id != name:
                    continue

                if rest:
                    # Found the object but don't know how to peek inside assignments
                    raise NodeNotFound(f'Found "{path}" but cannot go further')

                return node
            # Didn't find it in that assignment

        # Named code blocks
        elif type(node) in (ast.FunctionDef, ast.ClassDef):
            if node.name != name:
                continue

            # Found the object
            if not rest:
                return node

            # More to go, peek inside
            return find_name_in_nodes(rest, node.body, path)

    # Exhausted all nodes
    raise NodeNotFound(f'Couldn\'t find "{path}"')


def python_to_lineno(path, ref):
    """
    Look up a code reference in the specific python file and return the line number it
    is defined on

    Args:
        path (Path): Path to python file to examine
        ref (str): Variable, function or class name to find

    Returns:
        start (int): Line number the reference starts on
        end (int): Line number the reference ends on

    Raises:
        ParseError: If unable to resovle the reference for some reason
    """
    if path.suffix != ".py":
        raise ParseError("Source file is not Python")

    # Examine the file
    raw = path.read_text()

    return python_string_to_lineno(raw, ref)


def python_string_to_lineno(raw, ref):
    """
    Look up a code reference in the provided python source and return the line number it
    is defined on

    Args:
        filename (str): Path to python file to examine
        ref (str): Variable, function or class name to find

    Returns:
        start (int): Line number the reference starts on

    Raises:
        ParseError: If unable to resolve the reference for some reason
    """
    try:
        module = ast.parse(raw)
    except Exception as e:
        raise ParseError(str(e))

    try:
        node = find_name_in_nodes(ref, module.body)
    except NodeNotFound as e:
        raise ParseError(str(e))

    return node.lineno



"""
Remote definitions
"""
import re


if hasattr(re, "Pattern"):
    Pattern = re.Pattern
else:
    # Python 3.6
    Pattern = re._pattern_type


class Registry:
    """
    Remote registry
    """

    def __init__(self):
        self.remotes = []

    def register(self, cls):
        """
        Register the class
        """
        self.remotes.append(cls)
        return cls

    def get_by_url(self, url, branch):
        """
        Return an instantiated Remote for the given git Repo instance
        """
        for remote in self.remotes:
            for pattern in remote.get_remote_patterns():
                matches = pattern.match(url)
                if matches:
                    return remote(matches.group("repo"), branch)
        raise ValueError(f"Unable to find a match for {url}")


registry = Registry()


class Remote:
    """
    Base class for remotes
    """

    #: RegExp: Compiled regex pattern to match the remote url and extract the repo name.
    #: Must contain the named group ``repo``.
    remote_match = None

    #: str: Pattern of URL to code on remote site. Will be formatted with the values
    #: ``repo``, ``branch``, ``filename`` and ``line``
    url_pattern = None

    #: Pattern for line-based URLs. If a line number is specified, this will be rendered
    #: and passed to ``url_pattern`` as ``line``, otherwise ``line`` will be empty.
    url_pattern_line = None

    #: str: The identifying repo name on the remote site. Usually in the format
    #: ``username/repo``.
    repo = None

    #: str: The branch to link to.
    branch = None

    @classmethod
    def __init_subclass__(cls, **kwargs):
        """
        Automatically register Remote subclasses with the registry
        """
        super().__init_subclass__(**kwargs)
        registry.register(cls)

    @classmethod
    def get_remote_patterns(self):
        """
        Return remote_match as a list, in case it's defined as a single regex
        """
        if isinstance(self.remote_match, Pattern):
            return [self.remote_match]
        return self.remote_match

    def __init__(self, repo, branch):
        """
        Arguments:
            repo (str): The identifying repo name on the remote site. Usually in the
                format ``username/repo``.
            branch (str): The branch to link to.
        """
        self.repo = repo
        self.branch = branch

    def get_url(self, filename, line=None):
        """
        Return URL to file.

        Renders ``url_pattern`` and ``url_pattern_line``.
        """
        line_rendered = self.render_line(line)

        url = self.url_pattern.format(
            repo=self.repo, branch=self.branch, filename=filename, line=line_rendered
        )
        return url

    def render_line(self, line):
        if line is None:
            return ""
        return self.url_pattern_line.format(line=line)


class GitHub(Remote):
    """
    Repositories on https://github.com/
    """

    remote_match = [
        re.compile(r"^git@github.com:(?P<repo>.+?)\.git$"),
        re.compile(r"^https://github.com/(?P<repo>.+?)\.git$"),
    ]
    url_pattern = "https://github.com/{repo}/blob/{branch}/{filename}{line}"
    url_pattern_line = "#L{line}"


class Bitbucket(Remote):
    """
    Repositories on https://bitbucket.org/
    """

    remote_match = [
        re.compile(r"^git@bitbucket.org:(?P<repo>.+?)\.git$"),
        re.compile(r"^https://bitbucket.org/(?P<repo>.+?)\.git$"),
    ]
    url_pattern = "https://bitbucket.org/{repo}/src/{branch}/{filename}{line}"
    url_pattern_line = "#lines-{line}"


class GitLab(Remote):
    """
    Repositories on https://gitlab.com/
    """

    remote_match = [
        re.compile(r"^git@gitlab.com:(?P<repo>.+?)\.git$"),
        re.compile(r"^https://gitlab.com/(?P<repo>.+?)\.git$"),
    ]
    url_pattern = "https://gitlab.com/{repo}/blob/{branch}/{filename}{line}"
    url_pattern_line = "#L{line}"



"""
Set up the extension in sphinx

The ``setup()`` function will be called automatically by Sphinx
"""
from pathlib import Path

from .constants import DEFAULT_LABEL_FORMAT
from .git import Repo
from .remote import registry
from .role import gitref


def lookup_remote(app, config):
    """
    Once configuration is complete, build the Remote
    """
    try:
        remote_url = config.gitref_remote_url
        if not remote_url:
            raise AttributeError
    except AttributeError:
        raise ValueError("Could not determine gitref_remote_url, must set explicitly")

    try:
        branch = config.gitref_branch
        if not branch:
            raise AttributeError
    except AttributeError:
        raise ValueError("Could not determine gitref_branch, must set explicitly")

    config.gitref_remote = registry.get_by_url(remote_url, branch)


def setup(app):
    """
    Prepare config values and register role
    """
    # Pick up config defaults from current repo
    doc_root = Path(app.confdir)
    git_root = doc_root.parent / ".git"
    repo = Repo(git_root)

    # Add config variables
    app.add_config_value("gitref_remote_url", repo.get_remote_url(), "html")
    app.add_config_value("gitref_branch", repo.get_local_branch(), "html")
    app.add_config_value("gitref_label_format", DEFAULT_LABEL_FORMAT, "html")

    # Listen for config initialised
    app.connect("config-inited", lookup_remote)

    # Register role
    app.add_role("gitref", gitref)



"""
Working with git

This is a lightweight abstraction layer to provide access to necessary data from the
current git repository. At some point we may want the reliability improvement and
additional features available by moving to something like GitPython; that should be
possible with a drop-in replacement, but for now this should suffice.
"""
import re
from configparser import NoSectionError, RawConfigParser
from io import StringIO

from .constants import DEFAULT_REMOTE


class Repo:
    """
    Represent a local git repository
    """

    #: Path to ``.git`` repo
    path = None

    #: Name of the remote in ``.git/config``
    remote_name = None

    BRANCH_PATTERN = re.compile(r"^ref: refs/heads/(?P<branch>.+)$")

    def __init__(self, path, remote_name=DEFAULT_REMOTE):
        """
        Initialise a repo wrapper

        Args:
            path (Path): The path to the .git repo
        """
        if not path.is_dir():
            # If the pass doesn't exist then leave path as None so we can skip
            # unnecessary checks
            return

        self.path = path
        self.remote_name = remote_name

    def get_remote_url(self):
        """
        Find the URL for the current remote
        """
        if self.path is None:
            return None

        config_path = self.path / "config"
        if not config_path.is_file():
            return None

        # Read the config file and strip each line (configparser doesn't like tabs)
        config_src = "\n".join(
            [line.strip() for line in config_path.read_text().splitlines()]
        )
        config_io = StringIO(config_src)

        # Read the config - intentionally don't catch exceptions, we need those reported
        config = RawConfigParser(allow_no_value=True)
        config.readfp(config_io)

        try:
            url = config.get(f'remote "{self.remote_name}"', "url")
        except NoSectionError:
            return None

        return url

    def get_local_branch(self):
        """
        The current branch name
        """
        if self.path is None:
            return None

        # Get the current HEAD
        git_head_path = self.path / "HEAD"
        if not git_head_path.is_file():
            return None
        git_head = git_head_path.read_text().strip()

        # See if we're on a named branch
        matches = self.BRANCH_PATTERN.match(git_head)
        if matches:
            return matches.group("branch")

        # We're on a detached head, likely readthedocs; try to look it up in packed-refs
        packed_refs_path = self.path / "packed-refs"
        if not packed_refs_path.is_file():
            return None

        packed_refs_raw = packed_refs_path.read_text()
        remote_ref_prefix = f"refs/remotes/{self.remote_name}/"
        for line in packed_refs_raw.splitlines():
            # Skip irrelevant lines
            if line.startswith("#") or " " not in line:
                continue

            # Typical line:
            #   1234567890abcdef refs/remotes/origin/develop
            commit_hash, ref = line.split(" ", 1)
            if commit_hash == git_head and ref.startswith(remote_ref_prefix):
                ref = ref[len(remote_ref_prefix) :]
                return ref

        return None
